@@ -1,11 +1,10 @@
 import os
-from numpy.lib.npyio import load
 import six
 import copy
 import logging
 import numpy as np
 from dataclasses import dataclass, field, is_dataclass
-from typing import Optional, Union, List
+from typing import Dict, List, Optional, Union
 from functools import partial
 
 import trimesh
@@ -16,27 +15,85 @@ from lxml import etree
 _logger = logging.getLogger(__name__)
 
 
-@dataclass
+def _array_eq(arr1, arr2):
+    if arr1 is None and arr2 is None:
+        return True
+    return (
+        isinstance(arr1, np.ndarray)
+        and isinstance(arr2, np.ndarray)
+        and arr1.shape == arr2.shape
+        and (arr1 == arr2).all()
+    )
+
+
+@dataclass(eq=False)
 class TransmissionJoint:
     name: str
     hardware_interfaces: List[str] = field(default_factory=list)
 
+    def __eq__(self, other):
+        if not isinstance(other, TransmissionJoint):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and all(
+                self_hi in other.hardware_interfaces
+                for self_hi in self.hardware_interfaces
+            )
+            and all(
+                other_hi in self.hardware_interfaces
+                for other_hi in other.hardware_interfaces
+            )
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Actuator:
     name: str
     mechanical_reduction: Optional[float] = None
     # The follwing is only valid for ROS Indigo and prior versions
     hardware_interfaces: List[str] = field(default_factory=list)
 
+    def __eq__(self, other):
+        if not isinstance(other, Actuator):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and self.mechanical_reduction == other.mechanical_reduction
+            and all(
+                self_hi in other.hardware_interfaces
+                for self_hi in self.hardware_interfaces
+            )
+            and all(
+                other_hi in self.hardware_interfaces
+                for other_hi in other.hardware_interfaces
+            )
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Transmission:
     name: str
     type: Optional[str] = None
     joints: List[TransmissionJoint] = field(default_factory=list)
     actuators: List[Actuator] = field(default_factory=list)
 
+    def __eq__(self, other):
+        if not isinstance(other, Transmission):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and self.type == other.type
+            and all(self_joint in other.joints for self_joint in self.joints)
+            and all(other_joint in self.joints for other_joint in other.joints)
+            and all(
+                self_actuator in other.actuators for self_actuator in self.actuators
+            )
+            and all(
+                other_actuator in self.actuators for other_actuator in other.actuators
+            )
+        )
+
 
 @dataclass
 class Calibration:
@@ -70,16 +127,33 @@ class Cylinder:
     length: float
 
 
-@dataclass
+@dataclass(eq=False)
 class Box:
     size: np.ndarray
 
+    def __eq__(self, other):
+        if not isinstance(other, Box):
+            return NotImplemented
+        return _array_eq(self.size, other.size)
 
-@dataclass
+
+@dataclass(eq=False)
 class Mesh:
     filename: str
     scale: Optional[Union[float, np.ndarray]] = None
 
+    def __eq__(self, other):
+        if not isinstance(other, Mesh):
+            return NotImplemented
+
+        if self.filename != other.filename:
+            return False
+
+        if isinstance(self.scale, float) and isinstance(other.scale, float):
+            return self.scale == other.scale
+
+        return _array_eq(self.scale, other.scale)
+
 
 @dataclass
 class Geometry:
@@ -89,10 +163,15 @@ class Geometry:
     mesh: Optional[Mesh] = None
 
 
-@dataclass
+@dataclass(eq=False)
 class Color:
     rgba: np.ndarray
 
+    def __eq__(self, other):
+        if not isinstance(other, Color):
+            return NotImplemented
+        return _array_eq(self.rgba, other.rgba)
+
 
 @dataclass
 class Texture:
@@ -101,39 +180,85 @@ class Texture:
 
 @dataclass
 class Material:
+    name: Optional[str] = None
     color: Optional[Color] = None
     texture: Optional[Texture] = None
 
 
-@dataclass
+@dataclass(eq=False)
 class Visual:
     name: Optional[str] = None
     origin: Optional[np.ndarray] = None
     geometry: Optional[Geometry] = None  # That's not really optional according to ROS
     material: Optional[Material] = None
 
+    def __eq__(self, other):
+        if not isinstance(other, Visual):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and _array_eq(self.origin, other.origin)
+            and self.geometry == other.geometry
+            and self.material == other.material
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Collision:
     name: str
     origin: Optional[np.ndarray] = None
     geometry: Geometry = None
 
+    def __eq__(self, other):
+        if not isinstance(other, Collision):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and _array_eq(self.origin, other.origin)
+            and self.geometry == other.geometry
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Inertial:
     origin: Optional[np.ndarray] = None
     mass: Optional[float] = None
     inertia: Optional[np.ndarray] = None
 
+    def __eq__(self, other):
+        if not isinstance(other, Inertial):
+            return NotImplemented
+        return (
+            _array_eq(self.origin, other.origin)
+            and self.mass == other.mass
+            and _array_eq(self.inertia, other.inertia)
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Link:
     name: str
     inertial: Optional[Inertial] = None
     visuals: List[Visual] = field(default_factory=list)
     collisions: List[Collision] = field(default_factory=list)
 
+    def __eq__(self, other):
+        if not isinstance(other, Link):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and self.inertial == other.inertial
+            and all(self_visual in other.visuals for self_visual in self.visuals)
+            and all(other_visual in self.visuals for other_visual in other.visuals)
+            and all(
+                self_collision in other.collisions for self_collision in self.collisions
+            )
+            and all(
+                other_collision in self.collisions
+                for other_collision in other.collisions
+            )
+        )
+
 
 @dataclass
 class Dynamics:
@@ -149,7 +274,7 @@ class Limit:
     upper: Optional[float] = None
 
 
-@dataclass
+@dataclass(eq=False)
 class Joint:
     name: str
     type: str = None
@@ -163,15 +288,60 @@ class Joint:
     calibration: Optional[Calibration] = None
     safety_controller: Optional[SafetyController] = None
 
+    def __eq__(self, other):
+        if not isinstance(other, Joint):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and self.type == other.type
+            and self.parent == other.parent
+            and self.child == other.child
+            and _array_eq(self.origin, other.origin)
+            and _array_eq(self.axis, other.axis)
+            and self.dynamics == other.dynamics
+            and self.limit == other.limit
+            and self.mimic == other.mimic
+            and self.calibration == other.calibration
+            and self.safety_controller == other.safety_controller
+        )
 
-@dataclass
+
+@dataclass(eq=False)
 class Robot:
     name: str
     links: List[Link] = field(default_factory=list)
     joints: List[Joint] = field(default_factory=list)
+    materials: List[Material] = field(default_factory=list)
     transmission: List[str] = field(default_factory=list)
     gazebo: List[str] = field(default_factory=list)
 
+    def __eq__(self, other):
+        if not isinstance(other, Robot):
+            return NotImplemented
+        return (
+            self.name == other.name
+            and all(self_link in other.links for self_link in self.links)
+            and all(other_link in self.links for other_link in other.links)
+            and all(self_joint in other.joints for self_joint in self.joints)
+            and all(other_joint in self.joints for other_joint in other.joints)
+            and all(
+                self_material in other.materials for self_material in self.materials
+            )
+            and all(
+                other_material in self.materials for other_material in other.materials
+            )
+            and all(
+                self_transmission in other.transmission
+                for self_transmission in self.transmission
+            )
+            and all(
+                other_transmission in self.transmission
+                for other_transmission in other.transmission
+            )
+            and all(self_gazebo in other.gazebo for self_gazebo in self.gazebo)
+            and all(other_gazebo in self.gazebo for other_gazebo in other.gazebo)
+        )
+
 
 class URDFError(Exception):
     """General URDF exception."""
@@ -235,6 +405,34 @@ def _str2float(s):
     return float(s) if s is not None else None
 
 
+def apply_visual_color(
+    geom: trimesh.Trimesh,
+    visual: Visual,
+    material_map: Dict[str, Material],
+) -> None:
+    """Apply the color of the visual material to the mesh.
+
+    Args:
+        geom: Trimesh to color.
+        visual: Visual description from XML.
+        material_map: Dictionary mapping material names to their definitions.
+    """
+    if visual.material is None:
+        return
+
+    if visual.material.color is not None:
+        color = visual.material.color
+    elif visual.material.name is not None and visual.material.name in material_map:
+        color = material_map[visual.material.name].color
+    else:
+        return
+
+    if color is None:
+        return
+    if isinstance(geom.visual, trimesh.visual.ColorVisuals):
+        geom.visual.face_colors[:] = [int(255 * channel) for channel in color.rgba]
+
+
 def filename_handler_null(fname):
     """A lazy filename handler that simply returns its input.
 
@@ -256,11 +454,14 @@ def filename_handler_ignore_directive(fname):
     Returns:
         str: The file name without the prefix.
     """
-    return fname.split(f":{os.path.sep}{os.path.sep}")[-1]
+    if "://" in fname or ":\\\\" in fname:
+        return ":".join(fname.split(":")[1:])[2:]
+    return fname
 
 
 def filename_handler_ignore_directive_package(fname):
     """A filename handler that removes the 'package://' directive and the package it refers to.
+    It subsequently calls filename_handler_ignore_directive, i.e., it removes any other directive.
 
     Args:
         fname (str): A file name.
@@ -268,9 +469,10 @@ def filename_handler_ignore_directive_package(fname):
     Returns:
         str: The file name without 'package://' and the package name.
     """
-    if fname.startswith("package:"):
-        return os.path.sep.join(
-            fname.split(f":{os.path.sep}{os.path.sep}")[-1].split(os.path.sep)[1:]
+    if fname.startswith("package://"):
+        string_length = len("package://")
+        return os.path.join(
+            *os.path.normpath(fname[string_length:]).split(os.path.sep)[1:]
         )
     return filename_handler_ignore_directive(fname)
 
@@ -371,7 +573,10 @@ def filename_handler_magic(fname, dir):
     """
     return filename_handler_meta(
         fname=fname,
-        filename_handlers=[partial(filename_handler_relative, dir=dir)]
+        filename_handlers=[
+            partial(filename_handler_relative, dir=dir),
+            filename_handler_ignore_directive,
+        ]
         + _create_filename_handlers_to_urdf_file_recursive(urdf_fname=dir),
     )
 
@@ -633,7 +838,7 @@ class URDF:
         """Clear the validation error log."""
         self._errors = []
 
-    def show(self, collision_geometry=False):
+    def show(self, collision_geometry=False, callback=None):
         """Open a simpler viewer displaying the URDF model.
 
         Args:
@@ -645,14 +850,18 @@ class URDF:
                     "No collision scene available. Use build_collision_scene_graph=True and load_collision_meshes=True during loading."
                 )
             else:
-                self._scene_collision.show()
+                self._scene_collision.show(callback=callback)
         else:
             if self._scene is None:
                 raise ValueError(
                     "No scene available. Use build_scene_graph=True and load_meshes=True during loading."
                 )
+            elif len(self._scene.bounds_corners) < 1:
+                raise ValueError(
+                    "Scene is empty, maybe meshes failed to load? Use build_scene_graph=True and load_meshes=True during loading."
+                )
             else:
-                self._scene.show()
+                self._scene.show(callback=callback)
 
     def validate(self, validation_fn=None) -> bool:
         """Validate URDF model.
@@ -672,6 +881,10 @@ class URDF:
         return validation_fn(self._errors)
 
     def _create_maps(self):
+        self._material_map = {}
+        for m in self.robot.materials:
+            self._material_map[m.name] = m
+
         self._joint_map = {}
         for j in self.robot.joints:
             self._joint_map[j.name] = j
@@ -745,7 +958,7 @@ class URDF:
 
         try:
             parser = etree.XMLParser(remove_blank_text=True)
-            tree = etree.parse(fname_or_file, parser)
+            tree = etree.parse(fname_or_file, parser=parser)
             xml_root = tree.getroot()
         except Exception as e:
             _logger.error(e)
@@ -896,10 +1109,11 @@ class URDF:
         ]:
             matrix, joint_q = self._forward_kinematics_joint(j, q=q)
 
-            # update internal configuration vector - to enable mimic joint mapping
-            self._cfg[
-                self.actuated_dof_indices[self.actuated_joint_names.index(j.name)]
-            ] = joint_q
+            # update internal configuration vector - only consider actuated joints
+            if j.name in self.actuated_joint_names:
+                self._cfg[
+                    self.actuated_dof_indices[self.actuated_joint_names.index(j.name)]
+                ] = joint_q
 
             if self._scene is not None:
                 self._scene.graph.update(
@@ -957,7 +1171,7 @@ class URDF:
                 if g.geometry.mesh is not None:
                     if g.geometry.mesh.scale is not None:
                         S = np.eye(4)
-                        S[:3, :3] = np.diag(c.geometry.mesh.scale)
+                        S[:3, :3] = np.diag(g.geometry.mesh.scale)
                         pose = pose.dot(S)
                 m.apply_transform(pose)
                 meshes.append(m)
@@ -966,23 +1180,16 @@ class URDF:
         self._collision_mesh = meshes[0] + meshes[1:]
         return self._collision_mesh
 
-    def _geometry2trimeshscene(self, geometry, load_file, force_mesh):
+    def _geometry2trimeshscene(self, geometry, load_file, force_mesh, skip_materials):
         new_s = None
         if geometry.box is not None:
-            new_s = trimesh.Scene([trimesh.creation.box(extents=geometry.box.size)])
+            new_s = trimesh.primitives.Box(extents=geometry.box.size).scene()
         elif geometry.sphere is not None:
-            new_s = trimesh.Scene(
-                [trimesh.creation.uv_sphere(radius=geometry.sphere.radius)]
-            )
+            new_s = trimesh.primitives.Sphere(radius=geometry.sphere.radius).scene()
         elif geometry.cylinder is not None:
-            new_s = trimesh.Scene(
-                [
-                    trimesh.creation.cylinder(
-                        radius=geometry.cylinder.radius,
-                        height=geometry.cylinder.length,
-                    )
-                ]
-            )
+            new_s = trimesh.primitives.Cylinder(
+                radius=geometry.cylinder.radius, height=geometry.cylinder.length
+            ).scene()
         elif geometry.mesh is not None and load_file:
             new_filename = self._filename_handler(fname=geometry.mesh.filename)
 
@@ -990,7 +1197,12 @@ class URDF:
                 _logger.debug(f"Loading {geometry.mesh.filename} as {new_filename}")
 
                 if force_mesh:
-                    new_g = trimesh.load(new_filename, ignore_broken=True, force="mesh")
+                    new_g = trimesh.load(
+                        new_filename,
+                        ignore_broken=True,
+                        force="mesh",
+                        skip_materials=skip_materials,
+                    )
 
                     # add original filename
                     if "file_path" not in new_g.metadata:
@@ -1001,7 +1213,10 @@ class URDF:
                     new_s.add_geometry(new_g)
                 else:
                     new_s = trimesh.load(
-                        new_filename, ignore_broken=True, force="scene"
+                        new_filename,
+                        ignore_broken=True,
+                        force="scene",
+                        skip_materials=skip_materials,
                     )
 
                     if "file_path" in new_s.metadata:
@@ -1037,6 +1252,7 @@ class URDF:
         load_geometry,
         force_mesh,
         force_single_geometry,
+        skip_materials,
     ):
         if force_single_geometry:
             tmp_scene = trimesh.Scene(base_frame=link_name)
@@ -1049,13 +1265,18 @@ class URDF:
                     first_geom_name = v.name
 
                 new_s = self._geometry2trimeshscene(
-                    geometry=v.geometry, load_file=load_geometry, force_mesh=force_mesh
+                    geometry=v.geometry,
+                    load_file=load_geometry,
+                    force_mesh=force_mesh,
+                    skip_materials=skip_materials,
                 )
                 if new_s is not None:
                     origin = v.origin if v.origin is not None else np.eye(4)
 
                     if force_single_geometry:
                         for name, geom in new_s.geometry.items():
+                            if isinstance(v, Visual):
+                                apply_visual_color(geom, v, self._material_map)
                             tmp_scene.add_geometry(
                                 geometry=geom,
                                 geom_name=v.name,
@@ -1064,6 +1285,8 @@ class URDF:
                             )
                     else:
                         for name, geom in new_s.geometry.items():
+                            if isinstance(v, Visual):
+                                apply_visual_color(geom, v, self._material_map)
                             s.add_geometry(
                                 geometry=geom,
                                 geom_name=v.name,
@@ -1094,7 +1317,11 @@ class URDF:
             s.graph.update(frame_from=j.parent, frame_to=j.child, matrix=matrix)
 
         for l in self.robot.links:
-            s.graph.nodes.add(l.name)
+            if l.name not in s.graph.nodes and l.name != s.graph.base_frame:
+                _logger.warning(
+                    f"{l.name} not connected via joints. Will add link to base frame."
+                )
+                s.graph.update(frame_from=s.graph.base_frame, frame_to=l.name)
 
             meshes = l.collisions if use_collision_geometry else l.visuals
             self._add_geometries_to_scene(
@@ -1104,6 +1331,7 @@ class URDF:
                 load_geometry=load_geometry,
                 force_mesh=force_mesh,
                 force_single_geometry=force_single_geometry_per_link,
+                skip_materials=use_collision_geometry,
             )
 
         return s
@@ -1384,7 +1612,7 @@ class URDF:
         )
 
     def _parse_box(xml_element):
-        return Box(size=np.array(xml_element.attrib["size"].split()))
+        return Box(size=np.array(xml_element.attrib["size"].split(), dtype=np.float))
 
     def _write_box(self, xml_parent, box):
         etree.SubElement(
@@ -1559,7 +1787,7 @@ class URDF:
         if xml_element is None:
             return None
 
-        material = Material()
+        material = Material(name=xml_element.get("name"))
         material.color = URDF._parse_color(xml_element.find("color"))
         material.texture = URDF._parse_texture(xml_element.find("texture"))
 
@@ -1569,7 +1797,12 @@ class URDF:
         if material is None:
             return
 
-        xml_element = etree.SubElement(xml_parent, "material")
+        attrib = {"name": material.name} if material.name is not None else {}
+        xml_element = etree.SubElement(
+            xml_parent,
+            "material",
+            attrib=attrib,
+        )
 
         self._write_color(xml_element, material.color)
         self._write_texture(xml_element, material.texture)
@@ -1915,6 +2148,8 @@ class URDF:
             robot.links.append(URDF._parse_link(l))
         for j in xml_element.findall("joint"):
             robot.joints.append(URDF._parse_joint(j))
+        for m in xml_element.findall("material"):
+            robot.materials.append(URDF._parse_material(m))
         return robot
 
     def _validate_robot(self, robot):
@@ -1936,5 +2171,12 @@ class URDF:
             self._write_link(xml_element, link)
         for joint in robot.joints:
             self._write_joint(xml_element, joint)
+        for material in robot.materials:
+            self._write_material(xml_element, material)
 
         return xml_element
+
+    def __eq__(self, other):
+        if not isinstance(other, URDF):
+            raise NotImplemented
+        return self.robot == other.robot
